using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.DotProlog.InputOutput;
using OpcLabs.DotProlog.Parsing;
using OpcLabs.DotProlog.Parsing.Extensions;
using OpcLabs.DotProlog.Terms;

namespace DotPrologTest.Parsing
{
    public class ReadTermParserTestBase
    {
        #region Static Protected Methods

        static protected void TestError(
            [NotNull] string input,
            TermSyntaxErrorCode termSyntaxErrorCode)
        {
            TestError(
                input,
#pragma warning disable 618
                new TermParsingError(termSyntaxErrorCode));
#pragma warning restore 618
        }

        static protected void TestSuccess(
            [NotNull] string input,
            [NotNull] Term expectedTerm)
        {
            TestSuccess(input, new System.String[0], expectedTerm);
        }

        static protected void TestSuccess(
            [NotNull] string input, 
            [NotNull] IEnumerable<string> expectedVariableNames,
            [NotNull] Term expectedTerm)
        {
            TestSuccess(input, expectedVariableNames, expectedTerm, expectConsumed:true);
        }

        static protected void TestSuccess(
            [NotNull] string input,
            [NotNull] IEnumerable<string> expectedVariableNames,
            [NotNull] Term expectedTerm,
            bool expectConsumed)
        {
            Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Input: {0}", input));
            Trace.WriteLine("");

            var target = new ReadTermParser();
            TextStream textStream = TextStream.Create(input);
            var parsingContext = new ParsingContext();
            TermParsingError error = target.TryParse(textStream, ParsingParameters.Default, parsingContext, out Term term);
            Trace.WriteLine("");
            Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Error: {0}", error));
            Assert.IsNull(error, "Error is not null.");
            Assert.IsNotNull(term);
            Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Term: {0}", term));

            if (expectConsumed)
                Assert.AreEqual(-1, textStream.Get(), "Parsing has not consumed all the input.");
            else
                Assert.AreNotEqual(-1, textStream.Get(), "Parsing has consumed all the input.");

            var variableNameDictionary = parsingContext.VariableNameDictionary;
            List<string> expectedVariableNameList = expectedVariableNames.ToList();
            foreach (string variableName in expectedVariableNameList)
            {
                Debug.Assert(variableName != null);
                Assert.IsTrue(variableNameDictionary.ContainsKey(variableName));
            }
            Assert.AreEqual(expectedVariableNameList.Count, variableNameDictionary.Count);

            Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Exp.: {0}", expectedTerm));
            Solution solution = TermUtilities.UnifyWithoutOccursCheck(term, expectedTerm);   // IMPROVE: test structural equivalence instead
            Assert.IsNotNull(solution, "Unification failure.");
        }

        //static protected void TestUnconsumed([NotNull] string input)
        //{
        //    Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Input: {0}", input));
        //    Trace.WriteLine("");

        //    var target = new ReadTermParser();
        //    var textReader = new StringReader(input);
        //    Term term;
        //    var variableNameDictionary = new Dictionary<string, Variable>();
        //    TermParsingError error = target.TryParse(textReader, variableNameDictionary, out term);
        //    Trace.WriteLine("");
        //    Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Error: {0}", error));
        //    Assert.IsNull(error, "Error is not null.");
        //    Assert.IsNotNull(term);
        //    Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Term: {0}", term));

        //    Assert.AreNotEqual(-1, textReader.Read(), "Parsing has consumed all the input.");
        //}

        static protected void TestWait([NotNull] string input)
        {
            Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Input: {0}", input));
            Trace.WriteLine("");

            var target = new ReadTermParser();
            TextStream textStream = TextStream.Create(input);
            target.TryParse(textStream, ParsingParameters.Default, out Term term);

            Assert.AreEqual(-1, textStream.Get(), "Parsing has consumed all the input.");
        }

        #endregion



        #region Static Private Methods

        static private void TestError(
            [NotNull] string input,
            [NotNull] TermParsingError expectedError)
        {
            Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Input: {0}", input));
            Trace.WriteLine("");

            var target = new ReadTermParser();
            TextStream textStream = TextStream.Create(input);
            TermParsingError error = target.TryParse(textStream, ParsingParameters.Default, out Term term);
            Trace.WriteLine("");
            Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Term: {0}", term));
            Trace.WriteLine(System.String.Format(CultureInfo.InvariantCulture, "Error: {0}", error));
            Assert.IsNotNull(error, "Error was expected but not returned.");
            Assert.IsNull(term);

            Assert.AreEqual(expectedError, error);
        }

        #endregion
    }
}