using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.DotProlog.Code;
using OpcLabs.DotProlog.Execution;
using OpcLabs.DotProlog.Execution.Extensions;
using OpcLabs.DotProlog.InputOutput;
using OpcLabs.DotProlog.Parsing;
using OpcLabs.DotProlog.Terms;
using OpcLabs.DotProlog.Terms.Extensions;
using OpcLabs.UnitTestLib;

namespace OpcLabs.DotProlog.Test
{
    public class ExecutionTest
    {
        #region Static Public Methods

        [NotNull]
        static public IEnumerable<VariableDictionary> Failure()
        {
            return Enumerable.Empty<VariableDictionary>();
        }

        [NotNull]
        static public IEnumerable<VariableDictionary> Success()
        {
            return new[] { VariableDictionary.Empty };
        }

        [NotNull]
        static public IEnumerable<VariableDictionary> Success([NotNull] VariableDictionary variableDictionary)
        {
            return new[] { variableDictionary };
        }

        [NotNull]
        static public IEnumerable<VariableDictionary> Success(
            [NotNull] VariableDictionary variableDictionary1,
            [NotNull] VariableDictionary variableDictionary2)
        {
            return new[] { variableDictionary1, variableDictionary2 };
        }

        [NotNull]
        static public IEnumerable<VariableDictionary> Success([NotNull] Variable variable, [NotNull] Term value)
        {
            return Success(new VariableDictionary { { variable, value } });
        }

        [NotNull]
        static public IEnumerable<VariableDictionary> SuccessOrFailure(bool success)
        {
            return success ? Success() : Failure();
        }

        [NotNull]
        static public IEnumerable<VariableDictionary> SuccessOrFailure([CanBeNull] VariableDictionary variableDictionary)
        {
            return (variableDictionary != null) ? Success(variableDictionary) : Failure();
        }

        [NotNull]
        static public IEnumerable<VariableDictionary> SuccessOrFailure([NotNull] Variable variable, [CanBeNull] Term value)
        {
            return (value != null) ? Success(variable, value) : Failure();
        }

        #endregion



        #region Static Protected Methods

        [NotNull]
        static protected Term Parse([NotNull] string s) => Parse(s, out IDictionary<string, Variable> _);

        [NotNull]
        static protected Term Parse(
            [NotNull] string s, 
            [NotNull] out IDictionary<string, Variable> variableNameDictionary)
        {
            var readTermParser = new ReadTermParser();
            TextStream textStream = TextStream.Create(s);
            var parsingContext = new ParsingContext();
            TermParsingError error = readTermParser.TryParse(textStream, ParsingParameters.Default, parsingContext, out Term term);
            Debug.Assert(error == null);
            Debug.Assert(term != null);
            variableNameDictionary = parsingContext.VariableNameDictionary;
            return term;
        }

        [NotNull]
        static protected Clause ParseClause([NotNull] string s)
        {
            Term term = Parse(s);
            if (term.IsAtom || term.IsCompound)
            {
                if (term.Functor == new Functor(":-", 2))
                {
                    Term head = term.GetArgument(0);
                    Term body = term.GetArgument(1);
                    return new Clause(head, body);
                }
                return new Clause(term, CodeAtoms.True);
            }
            Debug.Assert(false);
            return null;
        }

        #endregion



        #region Public Methods

        public void Test(
            [NotNull] IEngine engine,
            [NotNull] Term goal,
            [NotNull] IEnumerable<VariableDictionary> expected)
        {
            Test(engine, goal, expected, -1);
        }

        public void Test(
            [NotNull] IEngine engine,
            [NotNull] Term goal,
            [NotNull] IEnumerable<VariableDictionary> expected,
            long maximumResultCount)
        {
            Test(engine, goal, expected, maximumResultCount, "No match.");
        }

        public void Test(
            [NotNull] IEngine engine,
            [NotNull] Term goal,
            [NotNull] IEnumerable<VariableDictionary> expected,
            [NotNull] string message)
        {
            Test(engine, goal, expected, maximumResultCount:-1, message);
        }

        public void Test(
            [NotNull] IEngine engine,
            [NotNull] Term goal,
            [NotNull] IEnumerable<VariableDictionary> expected,
            long maximumResultCount,
            [NotNull] string message)
        {
            // Arrange

            // Act
            IEnumerable<Solution> results = engine.Call(goal);

            // Assert
            bool match = expected.SequenceMatch(
                results.Select((solution, i) =>
                {
                    AssertUtilities.WriteLine($"Solution #{i}: {solution}");
                    return solution;
                }),
                (variableDictionary, solution) =>
                    // ReSharper disable PossibleNullReferenceException
                    // ReSharper disable AssignNullToNotNullAttribute
                    variableDictionary.Match(solution),
                // ReSharper restore AssignNullToNotNullAttribute
                // ReSharper restore PossibleNullReferenceException
                maximumResultCount);
            Assert.IsTrue(match, message);
        }

        public void TestError(
            [NotNull] IEngine engine,
            [NotNull] Term goal,
            [NotNull] Term errorTerm)
        {
            try
            {
                Test(engine, goal, Success());
            }
            catch (PrologException prologException)
            {
                Term ball = prologException.Ball;
                Trace.WriteLine(ball);
                Assert.IsNotNull(ball);
                var impDep = new Variable();
                Assert.IsNotNull(TermUtilities.UnifyWithoutOccursCheck(new NTerm("error") { errorTerm, impDep }, ball));
                return;
            }
            Assert.Fail();
        }

        public void TestException(
            [NotNull] IEngine engine,
            [NotNull] Term goal,
            [NotNull] Type exceptionType)
        {
            try
            {
                Test(engine, goal, Success());
            }
            catch (Exception exception)
            {
                Trace.WriteLine(exception.GetType());
                Trace.WriteLine(exception);
                Assert.IsInstanceOfType(exception, exceptionType);
                return;
            }
            Assert.Fail();
        }

        #endregion
    }
}