﻿using System;
using System.Diagnostics;
using System.IO;
using System.Reflection;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.IO.Extensions;
using OpcLabs.BaseLib.IO.Utilities;
using OpcLabs.BaseLib.Licensing;
using OpcLabs.BaseLib.Reflection.Extensions;
using OpcLabs.UnitTestLib;

namespace OpcLabs.BaseLibTest.Licensing
{
    public partial class LicenseMakerTest
    {
        #region Static Private Methods

        static private void Test(
            [NotNull] Func<LicenseMaker> createLicenseMakerFunction, 
            [NotNull] string grantFileName,
            [NotNull] string expectedLicenseFileName)
        {
            const string grantId = "Universal grant for creating QuickOPC 5.3+ licenses (1953000000-1999999999)";

            TestSection.Arrange();

            LicenseMaker licenseMaker = createLicenseMakerFunction();
            Debug.Assert(!(licenseMaker is null));

            var resourceLicenseStore = new ResourceLicenseStore();
            Assembly assembly = Assembly.GetExecutingAssembly();
            ResourceDescriptor grantResourceDescriptor =
                assembly.MakeNetResourceUri(typeof(LicenseMakerTest).Namespace + "." + grantFileName);
            resourceLicenseStore.AddLicenseResourceDescriptor(grantId, grantResourceDescriptor);
            licenseMaker.RegisterLicenseStore(resourceLicenseStore);

            Uri expectedLicenseBinaryUri =
                assembly.MakeNetResourceUri(typeof(LicenseMakerTest).Namespace + "." + expectedLicenseFileName);
            byte[] expectedLicenseBinary;
            using (Stream expectedLicenseBinaryStream =
                StreamUtilities.GetFromManagedResource(expectedLicenseBinaryUri, reflectionOnlyLoad: false))
                expectedLicenseBinary = expectedLicenseBinaryStream.ReadToArrayOfBytes();

            TestSection.Act();

            const string secretBase = "Johann Sebastian Bach";
            licenseMaker.CreateLicense(grantId, secretBase);

            licenseMaker.SetLicenseField("IssuerName", "OPC Labs");
            licenseMaker.SetLicenseField("NotAfter", (UInt64)0);
            licenseMaker.SetLicenseField("ReleaseDateCovered", (UInt64)132517728000000000);
            licenseMaker.SetLicenseField("SerialNumber", (UInt32)1999422743);
            licenseMaker.SetLicenseField("SubjectName", "OPC Labs Test");
            licenseMaker.SetLicenseField("Type", "Ultimate; Single Developer + Runtime Free License");
            licenseMaker.SetLicenseField("VersionNumberCovered", (UInt32)530);
            licenseMaker.SetLicenseField("OptionRestrictions", new[] { "-RtdServer", "-StreamInsight" });
            licenseMaker.SetLicenseField("AllPurposeStations", (UInt32)1);
            licenseMaker.SetLicenseField("ExtraRuntimeStations", unchecked((UInt32)(-1)));

            byte[] licenseBinary = licenseMaker.MakeLicenseBinary();
            licenseMaker.AbandonLicense();

            TestSection.Assert();
            Assert.IsNotNull(licenseBinary);
            int sequenceUnequalIndex = licenseBinary.SequenceUnequalIndex(expectedLicenseBinary);
            Assert.AreEqual(-1, sequenceUnequalIndex);
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void Test()
        {
            Test(
                createLicenseMakerFunction: () => new LicenseMaker(), 
                grantFileName: "0021000012.bin", 
                expectedLicenseFileName: "Key-Permanent-WebForm-1999422743-20201007.bin");
        }

        [TestMethod]
        public void Test_NotCppBinaryCompatible()
        {
            Test(
                createLicenseMakerFunction: () => new LicenseMaker { CppBinaryCompatible = false },
                grantFileName: "0021000012.bin",
                expectedLicenseFileName: "1999422743.bin");
        }

        #endregion
    }
}
