﻿using System;
using System.Collections.Generic;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib;
using OpcLabs.BaseLib.Instrumentation.Internal;
using OpcLabs.BaseLib.Timing;
using OpcLabs.EasyOpc.DataAccess;
using OpcLabs.EasyOpc.DataAccess.Implementation;
using OpcLabs.EasyOpc.DataAccess.Internal;
using OpcLabs.EasyOpc.DataAccess.OperationModel;
using OpcLabs.EasyOpc.DataAccess.Optimization.Implementation;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.DataAccess.Implementation
{
    [TestClass]
    public class ItemCachingEasyDAClientTest
    {
        [TestMethod]
        public void Test()
        {
            TestSection.Arrange();

            var apiMonitoringEntryList = new List<ApiMonitoringEntry>();

            // ReSharper disable once ParameterOnlyUsedForPreconditionCheck.Local
            void MonitorApiHandler(object sender, DataEventArgs<ApiMonitoringEntry> eventArgs, int level)
            {
                Assert.IsNotNull(sender);
                Assert.IsNotNull(eventArgs);
                ApiMonitoringEntry apiMonitoringEntry = eventArgs.Data;
                Assert.IsNotNull(apiMonitoringEntry);
                AssertUtilities.WriteLine($"({apiMonitoringEntryList.Count}) {new String(' ', level*4)}{apiMonitoringEntry}");
                apiMonitoringEntryList.Add(apiMonitoringEntry);
            }
            
            var simulatedClient = new SimulatedEasyDAClient();
            var simulatedClientApiMonitoringClient = new ApiMonitoringEasyDAClient(simulatedClient);
            simulatedClientApiMonitoringClient.MonitorApi += (sender, args) => MonitorApiHandler(sender, args, level: 1);

            var timing = new ManualTiming();

            var itemCachingClient = new ItemCachingEasyDAClient(simulatedClientApiMonitoringClient, timing);
            var target = new ApiMonitoringEasyDAClient(itemCachingClient);
            target.MonitorApi += (sender, args) => MonitorApiHandler(sender, args, level: 0);

            TestSection.Act();

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Read item 1 with default parameters.");
            apiMonitoringEntryList.Clear();
            var itemArguments1 = new DAItemArguments("s1", "i1");
            var readItemArguments1 = new DAReadItemArguments(itemArguments1);
            AssertUtilities.WriteLine($"{nameof(readItemArguments1)}: {readItemArguments1}");
            DAVtqResult[] resultArray1 = target.ReadMultipleItems(new []{readItemArguments1});
            Assert.IsNotNull(resultArray1);
            Assert.AreEqual(1, resultArray1.Length);
            DAVtqResult result1 = resultArray1[0];
            AssertUtilities.WriteLine($"{nameof(result1)}: {result1}");
            Assert.IsNotNull(result1);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(itemCachingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreEqual(4, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Read item 1 with default parameters again (cache hit).");
            apiMonitoringEntryList.Clear();
            DAVtqResult[] resultArray1B = target.ReadMultipleItems(new[] { readItemArguments1 });
            Assert.IsNotNull(resultArray1B);
            Assert.AreEqual(1, resultArray1B.Length);
            DAVtqResult result1B = resultArray1B[0];
            AssertUtilities.WriteLine($"{nameof(result1B)}: {result1B}");
            Assert.IsNotNull(result1B);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(itemCachingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreEqual(2, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Advance time to 500.");
            timing.AdvanceTo(500);

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Read item 1 without cache hit (too old).");
            apiMonitoringEntryList.Clear();
            DAVtqResult[] resultArray1C = target.ReadMultipleItems(new[] { readItemArguments1 });
            Assert.IsNotNull(resultArray1C);
            Assert.AreEqual(1, resultArray1C.Length);
            DAVtqResult result1C = resultArray1C[0];
            AssertUtilities.WriteLine($"{nameof(result1C)}: {result1C}");
            Assert.IsNotNull(result1C);
            Assert.AreEqual(result1, result1C);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(itemCachingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreEqual(4, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Read item 1 with cache hit.");
            apiMonitoringEntryList.Clear();
            var readItemArguments1D = new DAReadItemArguments(readItemArguments1) { ReadParameters = { ValueAge = 2000} };
            DAVtqResult[] resultArray1D = target.ReadMultipleItems(new[] { readItemArguments1D });
            Assert.IsNotNull(resultArray1D);
            Assert.AreEqual(1, resultArray1D.Length);
            DAVtqResult result1D = resultArray1D[0];
            AssertUtilities.WriteLine($"{nameof(result1D)}: {result1D}");
            Assert.IsNotNull(result1D);
            Assert.AreEqual(result1, result1D);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(itemCachingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreEqual(2, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Read item 1 from the device.");
            apiMonitoringEntryList.Clear();
            var readItemArguments1E = new DAReadItemArguments(readItemArguments1) {ReadParameters = {DataSource = DADataSource.Device}};
            AssertUtilities.WriteLine($"{nameof(readItemArguments1E)}: {readItemArguments1E}");
            DAVtqResult[] resultArray1E = target.ReadMultipleItems(new[] { readItemArguments1E});
            Assert.IsNotNull(resultArray1E);
            Assert.AreEqual(1, resultArray1E.Length);
            DAVtqResult result1E = resultArray1E[0];
            AssertUtilities.WriteLine($"{nameof(result1E)}: {result1E}");
            Assert.IsNotNull(result1E);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(itemCachingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreEqual(4, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Subscribe to item 2.");
            apiMonitoringEntryList.Clear();
            EasyDAItemSubscriptionArguments itemSubscriptionArguments2 = new EasyDAItemSubscriptionArguments("s2", "i2", 2000, "ItemSubscription2");
            AssertUtilities.WriteLine($"{nameof(itemSubscriptionArguments2)}: {itemSubscriptionArguments2}");
            int handle2 = target.SubscribeItem(itemSubscriptionArguments2);
            AssertUtilities.WriteLine($"{nameof(handle2)}: {handle2}");

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Read item 2 with cache hit.");
            apiMonitoringEntryList.Clear();
            var itemArguments2 = new DAItemArguments("s2", "i2");
            var readItemArguments2A = new DAReadItemArguments(itemArguments2) { ReadParameters = {ValueAge = 3000}};
            AssertUtilities.WriteLine($"{nameof(readItemArguments2A)}: {readItemArguments2A}");
            DAVtqResult[] resultArray2A = target.ReadMultipleItems(new[] { readItemArguments2A });
            Assert.IsNotNull(resultArray2A);
            Assert.AreEqual(1, resultArray2A.Length);
            DAVtqResult result2A = resultArray2A[0];
            AssertUtilities.WriteLine($"{nameof(result2A)}: {result2A}");
            Assert.IsNotNull(result2A);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(itemCachingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreEqual(2, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException

            AssertUtilities.WriteLine();
            AssertUtilities.WriteLine("Read item 2 with default parameters (no cache hit).");
            apiMonitoringEntryList.Clear();
            var readItemArguments2B = new DAReadItemArguments(itemArguments2);
            AssertUtilities.WriteLine($"{nameof(readItemArguments2B)}: {readItemArguments2B}");
            DAVtqResult[] resultArray2B = target.ReadMultipleItems(new[] { readItemArguments2B });
            Assert.IsNotNull(resultArray2B);
            Assert.AreEqual(1, resultArray2B.Length);
            DAVtqResult result2B = resultArray2B[0];
            AssertUtilities.WriteLine($"{nameof(result2B)}: {result2B}");
            Assert.IsNotNull(result2B);
            // ReSharper disable PossibleNullReferenceException
            Assert.AreSame(itemCachingClient, apiMonitoringEntryList[0].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[0].TypedMemberName);
            Assert.AreSame(simulatedClient, apiMonitoringEntryList[1].TargetOrSourceObject);
            Assert.AreEqual("IEasyDAClient.ReadMultipleItems", apiMonitoringEntryList[1].TypedMemberName);
            Assert.AreEqual(4, apiMonitoringEntryList.Count);
            // ReSharper restore PossibleNullReferenceException
        }
    }
}
